
#include <stdio.h>
#include <vector>
#include <fstream>
#include <cstdlib>
#include <cstring>
#include <string>
#include <math.h>
#include <ctime>
#include <opencv/cv.h>
#include <opencv/cvaux.h>
#include <opencv/highgui.h> 
#include <GL/glut.h>
#include <pthread.h>
#include "yaml-cpp/yaml.h"
#include "muParser.h"
// =======================
// YAGSBPL libraries
#include "yagsbpl/yagsbpl_base.h"
#include "yagsbpl/planners/A_star.h"
#include "glt_zpr/zpr.h"
// =======================
// local libraries
#include "map2D.h"
#include "cvAux2.h"

#define OPENGL_WIN_SIZE 500
#define CTRLLIM 0.38268343236508984
#define __max(x,y) (((x)>(y))?(x):(y))
#define POLE_CUTOFF 10

//pthread_mutex_t mutexGlobal = PTHREAD_MUTEX_INITIALIZER;

// Helper class:
class ptXY { public: double x, y; };
ptXY createPtXY(double xx=0.0, double yy=0.0) { ptXY tmppt; tmppt.x=xx; tmppt.y=yy; return (tmppt); }


// Global variables
double dx, dy, ox, oy;
std::vector<int> px, py, rr2;
std::vector<int> hx, hy;
Map_t<int> obs_map;
Map_t<double> weight_map;
Map_t<double> metric_map; // contains square-root of the determinant of metric tensor (\zeta)
Map_t<double> metric_mat_map[2][2];
Map_t<int> unknownobs_map;
Map_t<double> probsum_map, entropy_map, smooth_entropy_map; Map_t<int> probcount_map; // Occupancy  probability information for exploration problem
Map_t<int> power_voronoi_map;
Map_t<int> explore_contract_map;
double weightMinLim, weightMaxLim, weightLimRange;
double metricMinLim, metricMaxLim, metricLimRange;
double maxPowerDist, minPowerDist;
std::vector<ptXY> controlIntegral;
std::vector<double> controlAbsIntegral;
std::vector<int> aboveThreshPtX; std::vector<int> aboveThreshPtY;
std::vector<int> imgFrames;
double minWeight=1e10, maxWeight=-1e10, maxMinWeight;
double minMetric=1e10, maxMetric=-1e10, maxMinMetric;
bool explore=false, altermetric=false, allowusercontrol=false;
int unknownobsthresh, entropysmoothenradius;
double obstacleprobthresh, freezeentropy, thresholdentropy, tinyweight, tinymetric, controlmagzero;
int userrobot;
int iter;

bool explicitMetricMat=false;
mu::Parser metricMat[2][2];
double mmx, mmy;
std::vector< std::vector<double> > geodesic_cotangents[2]; // geodesic_cotangents[i][n][e] - i=0,1 ; e=0..7; n=0...(N-1)

GLvoid* imageData;
int WIDTH, HEIGHT;
// ========================================================
// Parse the environment file

// ========================================================
// Comptibility with older version of YAML-cpp

template <class T>
void operator >> (const YAML::Node& node, T& v) {
    v  = node.as<T>();
}

#define FindValue(s)  operator[](s)

// ========================================================

double rectPulse(double x, double a, double b)
{
    if (a<=x && x<b) return (1.0);
    else return (0.0);
}

void parseEnvFile(const char* fName)
{
	std::ifstream fin(fName);
	/*YAML::Parser parser(fin);
	YAML::Node doc;
	parser.GetNextDocument(doc); */
	YAML::Node doc = YAML::LoadFile (fName);
	
	std::string obsmapFileName, weightmapFileName, metricmapFileName, unknownobsmapFileName;
	IplImage* image_obsmap; IplImage* image_weightmap; IplImage* image_metricmap; IplImage* image_unknownobsmap; 
	int obsthresh;
	int obsPixel, weightPixel, metricPixel, unknownobsPixel;
	double sx, sy, rad;
	
	double weightScaleFactor;
	bool invertweight;
	mu::Parser weightFun;
	double wfx, wfy;
	bool explicitWightFun = false;
	
	double metricScaleFactor;
	bool invertmetric;
	mu::Parser metricFun;
	double mfx, mfy;
	bool explicitMetricFun=false;
	double metricMatNorm;
	
	if (doc.FindValue("weightfunction")) {
	    explicitWightFun = true;
	    std::string expr; doc["weightfunction"] >> expr;
	    weightFun.DefineVar("x", &wfx); weightFun.DefineVar("y", &wfy);
	    weightFun.DefineFun("rect", rectPulse, false);
	    weightFun.SetExpr(expr);
	}
	
	if (doc.FindValue("metricfunction")) {
	    explicitMetricFun = true;
	    std::string expr; doc["metricfunction"] >> expr;
	    metricFun.DefineVar("x", &mfx); metricFun.DefineVar("y", &mfy);
	    metricFun.DefineFun("rect", rectPulse, false);
	    metricFun.SetExpr(expr);
	}
	else if (doc.FindValue("metricmatrix")) {
	    explicitMetricMat = true;
	    for (int a=0; a<2; a++)
	        for (int b=0; b<2; b++) {
	            std::string expr; doc["metricmatrix"][a][b] >> expr;
	            metricMat[a][b].DefineVar("x", &mmx); metricMat[a][b].DefineVar("y", &mmy);
	            metricMat[a][b].SetExpr(expr);
	        }
	}
	
	if (doc.FindValue("explore")) {
	    doc["explore"] >> explore;
	    if (explore) {
	        doc["unknownobsmap"] >> unknownobsmapFileName;
	        doc["unknownobsthresh"] >> unknownobsthresh;
	        doc["obstacleprobthresh"] >> obstacleprobthresh;
	        doc["altermetric"] >> altermetric;
	        doc["freezeentropy"] >> freezeentropy;
	        doc["thresholdentropy"] >> thresholdentropy;
	        doc["tinyweight"] >> tinyweight;
	        doc["tinymetric"] >> tinymetric;
	        doc["controlmagzero"] >> controlmagzero;
	        doc["entropysmoothenradius"] >> entropysmoothenradius;
	        if (explicitWightFun && altermetric) {
	            printf("WARNING: Ignoring metric matrix since metric alteration is needed by exploration.");
	            explicitWightFun = false;
	        }
	    }
	}
	
	// ----------------------------------------
	doc["obsmap"] >> obsmapFileName;
	doc["weightmap"] >> weightmapFileName;
	doc["metricmap"] >> metricmapFileName;
	
	doc["obsthresh"] >> obsthresh;
	doc["normalizeweight"][0] >> weightMinLim;
	doc["normalizeweight"][1] >> weightMaxLim;
	doc["invertweight"] >> invertweight;
	if ( doc.FindValue("normalizemetric") && doc.FindValue("invertmetric") ) {
        doc["normalizemetric"][0] >> metricMinLim;
        doc["normalizemetric"][1] >> metricMaxLim;
        doc["invertmetric"] >> invertmetric;
    }
	
	printf("Loading map: %s\n", obsmapFileName.c_str());
	image_obsmap = cvLoadImage( obsmapFileName.c_str(), CV_LOAD_IMAGE_GRAYSCALE );
	printf("Loading map: %s\n", weightmapFileName.c_str());
	image_weightmap = cvLoadImage( weightmapFileName.c_str(), CV_LOAD_IMAGE_GRAYSCALE );
	printf("Loading map: %s\n", metricmapFileName.c_str());
	image_metricmap = cvLoadImage( metricmapFileName.c_str(), CV_LOAD_IMAGE_GRAYSCALE );
	if (explore) {
	    printf("Loading map: %s\n", unknownobsmapFileName.c_str());
	    image_unknownobsmap = cvLoadImage( unknownobsmapFileName.c_str(), CV_LOAD_IMAGE_GRAYSCALE );
	}
	
	if (doc.FindValue("cellsize")) {
	    doc["cellsize"][0] >> dx;
	    doc["cellsize"][1] >> dy;
	}
	else {
	    double imgWidth, imgHeight;
	    doc["dimensions"][0] >> imgWidth;
	    doc["dimensions"][1] >> imgHeight;
	    dx = imgWidth / (double)image_obsmap->width;
	    dy = imgHeight / (double)image_obsmap->height;
	}
	doc["origin"][0] >> ox;
	doc["origin"][1] >> oy;
	
	// Create the maps using image
	obs_map.init(image_obsmap->width, image_obsmap->height, 0);
	weight_map.init(image_weightmap->width, image_weightmap->height, weightMaxLim);
	
	if (explicitMetricMat) {
	    for (int a=0; a<2; a++)
	        for (int b=0; b<2; b++) {
	            metric_mat_map[a][b].init(image_metricmap->width, image_metricmap->height, ((a==b)?1.0:0.0) );
	            metric_map.init(image_metricmap->width, image_metricmap->height, 1.0); // stores the determinant
	        }
	}
	else
    	metric_map.init(image_metricmap->width, image_metricmap->height, metricMaxLim);
    
    if (explore) {
        unknownobs_map.init(image_unknownobsmap->width, image_unknownobsmap->height, 0);
        probsum_map.init(image_unknownobsmap->width, image_unknownobsmap->height, 0.5);
        probcount_map.init(image_unknownobsmap->width, image_unknownobsmap->height, 1);
        explore_contract_map.init(image_unknownobsmap->width, image_unknownobsmap->height, 0);
        entropy_map.init(image_unknownobsmap->width, image_unknownobsmap->height, 0.0);
        smooth_entropy_map.init(image_unknownobsmap->width, image_unknownobsmap->height, 0.0);
    }
    
	printf("Image size: %d by %d. widthStep: %d. \n", obs_map.x_size, obs_map.y_size, image_obsmap->widthStep);
	if( image_obsmap->width!=image_weightmap->width || image_weightmap->width!=image_metricmap->width || image_obsmap->height!=image_weightmap->height || image_weightmap->height!=image_metricmap->height ) {
	    printf("ERROR: Image sizes do not match!");
	    exit(1);
	}
	
	for (int a=0; a<obs_map.x_size; a++)
        for (int b=0; b<obs_map.y_size; b++) {
            //(int) (((uchar *)(image_original->imageData + (image_original->height-b-1)*image_original->widthStep))[a]);
            
            weightPixel = getCVpixel(image_weightmap,a,b);
            if (explicitWightFun) {
                wfx = ((double)a)*dx - ox;
                wfy = ((double)b)*dy - oy;
                weight_map[a][b] = weightFun.Eval();
            }
            else
                weight_map[a][b] = (double)weightPixel;
            
            if (explicitMetricMat) {
                mmx = ((double)a)*dx - ox;
                mmy = ((double)b)*dy - oy;
                for (int c=0; c<2; c++)
	                for (int d=0; d<2; d++) 
	                    metric_mat_map[c][d][a][b] = metricMat[c][d].Eval();
	            metric_map[a][b] = sqrt( metric_mat_map[0][0][a][b]*metric_mat_map[1][1][a][b] - metric_mat_map[1][0][a][b]*metric_mat_map[0][1][a][b] );
            }
            else {
                metricPixel = getCVpixel(image_metricmap,a,b);
                if (explicitMetricFun) {
                    mfx = ((double)a)*dx - ox;
                    mfy = ((double)b)*dy - oy;
                    metric_map[a][b] = metricFun.Eval();
                }
                else
                    metric_map[a][b] = (double)metricPixel;
            }
            
            obsPixel = getCVpixel(image_obsmap,a,b);
            if (obsPixel <= obsthresh)
                obs_map[a][b] = 1;
            else {
                if (weight_map[a][b] < minWeight) minWeight = weight_map[a][b];
                if (weight_map[a][b] > maxWeight) maxWeight = weight_map[a][b];
                //if (!explicitMetricMat) {
                    if (metric_map[a][b] < minMetric) minMetric = metric_map[a][b];
                    if (metric_map[a][b] > maxMetric) maxMetric = metric_map[a][b];
                /* }
                else {
                    metricMatNorm = sqrt( pow(metric_mat_map[0][0][a][b],2.0) + pow(metric_mat_map[0][1][a][b],2.0) + pow(metric_mat_map[1][0][a][b],2.0) + pow(metric_mat_map[1][1][a][b],2.0) );
                    if (metricMatNorm < minMetric) minMetric = metricMatNorm;
                    if (metricMatNorm > maxMetric) maxMetric = metricMatNorm;
                } */
            }
            
            if (explore) {
                unknownobsPixel = getCVpixel(image_unknownobsmap,a,b);
                if (unknownobsPixel <= unknownobsthresh)
                    unknownobs_map[a][b] = 1;
            }
        }
    printf("Weight values: Min %0.3f - Max %0.3f\n", minWeight, maxWeight);
    printf("Metric values: Min %0.3f - Max %0.3f\n", minMetric, maxMetric);
    maxMinWeight = maxWeight + minWeight;
    maxMinMetric = maxMetric + minMetric;
    
    if (invertweight)
	    for (int a=0; a<obs_map.x_size; a++)
            for (int b=0; b<obs_map.y_size; b++) 
                weight_map[a][b] = maxMinWeight - weight_map[a][b];
    weightLimRange = weightMaxLim - weightMinLim;
    if (maxWeight-minWeight > 0.0) {
        weightScaleFactor = weightLimRange / (maxWeight - minWeight);
        for (int a=0; a<obs_map.x_size; a++)
            for (int b=0; b<obs_map.y_size; b++) 
                weight_map[a][b] = weightScaleFactor * (weight_map[a][b] - minWeight)  +  weightMinLim;
    }
    
    if (!explicitMetricMat) {
        if (invertmetric)
	        for (int a=0; a<obs_map.x_size; a++)
                for (int b=0; b<obs_map.y_size; b++) 
                    metric_map[a][b] = maxMinMetric - metric_map[a][b];
        metricLimRange = metricMaxLim - metricMinLim;
        if (maxMetric-minMetric > 0.0) {
            metricScaleFactor = metricLimRange / (maxMetric - minMetric);
            //printf("Metric scale: %f, %f, %f \n", maxMetric, minMetric, metricScaleFactor);
            for (int a=0; a<metric_map.x_size; a++)
                for (int b=0; b<metric_map.y_size; b++) {
                    metric_map[a][b] = metricScaleFactor * (metric_map[a][b] - minMetric)  +  metricMinLim;
                    //printf("%d,%d,%f; ", a, b, metric_map[a][b]);
                }
        }
    }
	
	for (int a=0; a<doc["starts"].size(); a++) {
	    doc["starts"][a][0] >> sx;
	    doc["starts"][a][1] >> sy;
	    doc["radius"][a] >> rad;
	    px.push_back((int)((sx+ox)/dx));
	    py.push_back((int)((sy+oy)/dy));
	    rr2.push_back((int)(rad*rad/(dx*dx+dy*dy)));
	}
	
	userrobot = -1;
	if (doc.FindValue("allowusercontrol")) 
	    doc["allowusercontrol"] >> allowusercontrol;
	
	if (doc.FindValue("imgframes")) 
	    for (int a=0; a<doc["imgframes"].size(); a++)
	        imgFrames.push_back(doc["imgframes"][a].as<int>());
	
}

// ===========================================================
// Classes and functions used by YAGSBPL

class myNode
{
public:
	int x, y;
	int SeedIndex, SeedNeighbourX, SeedNeighbourY;
	double SeedNeighbourWeight; // TODO: Add this to the other main.
	
	//bool isVirtual;
	//int id; // -1 NP, xMax SP, integer rest
	int pole; // 0 for regular vertices, 1 for NP, -1 for SP
	
	bool operator==(const myNode& n) { 
	    if ( !(n.pole || pole) )
    	    return (x==n.x && y==n.y); 
    	else if (n.pole && pole)
	        return (pole==n.pole);
    	else
    	    return (false);
	}; // This must be defined for the node
	
	void print(void) { 
	    if (pole)
	        printf("(%d) ", pole);
	    else
    	    printf("(%d,%d) ", x, y); 
	}
};


A_star_planner<myNode,double>* planner;

/* myNode createMyNode (int x, int y) {
    myNode tempNode;
    tempNode.x = x; tempNode.y = y;
    return (tempNode);
} */

// -----

void event_SuccUpdated(myNode &n, myNode &nn, double &edgeCost, double &gVal, double &fVal, int &seedLineage)
{
    if ( !n.pole  &&  n.x==px[n.SeedIndex]  &&  n.y==py[n.SeedIndex] ) { // TODO virtual: assign nn.SeedNeighbourX & nn.SeedNeighbourY correctly.
        /* if (n.x==0  &&  nn.pole==1) { // connection across north pole
            nn.SeedNeighbourX = -1;
            nn.SeedNeighbourY = 0;
            nn.SeedNeighbourWeight = 0.333; // TODO: Add this to the other main.
        }
        else if (n.x==obs_map.x_size-1  &&  nn.pole==-1) { // connection across south pole
            nn.SeedNeighbourX = 1;
            nn.SeedNeighbourY = 0;
            nn.SeedNeighbourWeight = 0.333; // TODO: Add this to the other main.
        }
        else PolesRemoved */ if (n.y==0  &&  nn.y==obs_map.y_size-1) { // connection across meridian
            nn.SeedNeighbourX = nn.x - n.x;
            nn.SeedNeighbourY = -1;
            nn.SeedNeighbourWeight = 1.0; // TODO: Add this to the other main.
        }
        else if (n.y==obs_map.y_size-1  &&  nn.y==0) {  // connection across meridian
            nn.SeedNeighbourX = nn.x - n.x;
            nn.SeedNeighbourY = 1;
            nn.SeedNeighbourWeight = 1.0; // TODO: Add this to the other main.
        }
        else {
            nn.SeedNeighbourX = nn.x - n.x;
            nn.SeedNeighbourY = nn.y - n.y;
            nn.SeedNeighbourWeight = 1.0; // TODO: Add this to the other main.
        }
    }
    else {
        nn.SeedNeighbourX = n.SeedNeighbourX;
        nn.SeedNeighbourY = n.SeedNeighbourY;
        nn.SeedNeighbourWeight = n.SeedNeighbourWeight; // TODO: Add this to the other main.
    }
    
    nn.SeedIndex = n.SeedIndex;
    
    if (explore) {
        if ( explore_contract_map[n.x][n.y] && entropy_map[nn.x][nn.y]<=1.001*freezeentropy ) //weight_map[nn.x][nn.y]<=1.01*tinyweight ) 
            explore_contract_map[nn.x][nn.y] = explore_contract_map[n.x][n.y];
    }
}


int expandCount1, expandCountTot;

void event_NodeExpanded(myNode n, double gVal, double fVal, int seedLineage)
{
    double seedNeighborEdgeLen = sqrt( (double)n.SeedNeighbourX*(double)n.SeedNeighbourX + (double)n.SeedNeighbourY*(double)n.SeedNeighbourY );
    if (seedNeighborEdgeLen==0.0) seedNeighborEdgeLen=1.0;
        //printf("%f,%f,%f ", (double)n.SeedNeighbourX, (double)n.SeedNeighbourY, seedNeighborEdgeLen);
    //controlIntegral[n.SeedIndex].x += gVal*weight_map[n.x][n.y]*((double)n.SeedNeighbourX)/seedNeighborEdgeLen;
    //controlIntegral[n.SeedIndex].y += gVal*weight_map[n.x][n.y]*((double)n.SeedNeighbourY)/seedNeighborEdgeLen;
    expandCountTot++;
    
    if (n.pole)
        return;
    
    if (explicitMetricMat) {
        int e=0;
        for (int a=-1; a<=1; a++)
		    for (int b=-1; b<=1; b++) {
			    if (a==0 && b==0) continue;
			    if (n.SeedNeighbourX==a && n.SeedNeighbourY==b) {
                    //geodesic_cotangents[i][n][e]
                    controlIntegral[n.SeedIndex].x += gVal * weight_map[n.x][n.y] * geodesic_cotangents[0][n.SeedIndex][e] * metric_map[n.x][n.y] 
                                                        * n.SeedNeighbourWeight; // TODO: Add this to the other main.
                    controlIntegral[n.SeedIndex].y += gVal * weight_map[n.x][n.y] * geodesic_cotangents[1][n.SeedIndex][e] * metric_map[n.x][n.y] 
                                                        * n.SeedNeighbourWeight; // TODO: Add this to the other main.
                    // debug
                    /* if (n.SeedIndex==1 && px[1]==obs_map.x_size-1   &&   fabs(gVal * weight_map[n.x][n.y] * geodesic_cotangents[0][n.SeedIndex][e] * metric_map[n.x][n.y] * n.SeedNeighbourWeight)>0 ) {
                        printf("%f,%f; ", gVal * weight_map[n.x][n.y] * geodesic_cotangents[0][n.SeedIndex][e] * metric_map[n.x][n.y] * n.SeedNeighbourWeight, gVal * weight_map[n.x][n.y] * geodesic_cotangents[1][n.SeedIndex][e] * metric_map[n.x][n.y] * n.SeedNeighbourWeight);
                        expandCount1++;
                    } */
                }
                e++;
            }
    }
    else {
        controlIntegral[n.SeedIndex].x += gVal*weight_map[n.x][n.y]* metric_map[n.x][n.y] *((double)n.SeedNeighbourX)  
                                                        * n.SeedNeighbourWeight / seedNeighborEdgeLen; // TODO: Add this to the other main.
        controlIntegral[n.SeedIndex].y += gVal*weight_map[n.x][n.y]* metric_map[n.x][n.y] *((double)n.SeedNeighbourY)  
                                                        * n.SeedNeighbourWeight/ seedNeighborEdgeLen; // TODO: Add this to the other main.
    }
    
    if (explore) {
        controlAbsIntegral[n.SeedIndex] += sqrt(controlIntegral[n.SeedIndex].x*controlIntegral[n.SeedIndex].x + controlIntegral[n.SeedIndex].y*controlIntegral[n.SeedIndex].y);
        
        if ( aboveThreshPtX[n.SeedIndex] < 0 && smooth_entropy_map[n.x][n.y] > thresholdentropy ){
            aboveThreshPtX[n.SeedIndex] = n.x;
            aboveThreshPtY[n.SeedIndex] = n.y;
        }
    }
}

// -----

int getHashBin(myNode& n) // Use the absolute value of x coordinate as hash bin counter. Not a good choice though!
{
    if (n.pole)
        return (1);
    else
    	return (n.x+2);
}

bool isAccessible(myNode& n)
{
    if (n.pole)
        return (true);
    else
    	return ( n.x>=0  &&  n.x<obs_map.x_size  &&  n.y>=0  &&  n.y<obs_map.y_size   &&   !obs_map[n.x][n.y] );
	//return ( !obs_map[n.x][n.y] );
}

int CurrentRobIndex;
bool isInTesellation(myNode& n)
{
	return ( power_voronoi_map[n.x][n.y] == CurrentRobIndex );
}

Map_t<double> DistMap3x3;
double Dx[2];
double csum;

void getSuccessors(myNode& n, std::vector<myNode>* s, std::vector<double>* c) // Define a 8-connected graph
{
	// This function needn't account for obstacles or size of environment. That's done by "isAccessible"
	myNode tn;
	s->clear(); c->clear(); // Planner is supposed to clear these. Still, for safety we clear it again.
	
	
	/* if (n.pole==1) { // north pole
	    tn.x = 0;
	    tn.pole = 0;
	    for (int a=0; a<obs_map.y_size; a++) {
	        tn.y = a;
	        s->push_back(tn); c->push_back(1); // 1e-10
	        //tn.print();
	    }
	    return;
	}
	
	if (n.pole==-1) { // south pole
	    tn.x = obs_map.x_size - 1;
	    tn.pole = 0;
	    for (int a=0; a<obs_map.y_size; a++) {
	        tn.y = a;
	        s->push_back(tn); c->push_back(1); // 1e-10
	        //tn.print();
	    }
	    return;
	} PolesRemoved */
	
	// normal vertex
	for (int a=-1; a<=1; a++)
		for (int b=-1; b<=1; b++) {
			if (a==0 && b==0) continue;
			tn.x = n.x + a;
			tn.y = n.y + b;
			tn.pole = 0;

            /* if (tn.x<0 && b==0) { // connection across north pole
                //tn.x = 0;
                //tn.y = ( tn.y + (int)(obs_map.y_size/2) ) % obs_map.y_size; // best if obs_map.y_size is even.
                tn.pole = 1;
                s->push_back(tn); c->push_back(1); // 1e-10
                //printf(" In north pole:");
                //tn.print();
                continue;
            }
            else if (tn.x>=obs_map.x_size && b==0) { // connection across south pole
                //tn.x = obs_map.x_size - 1;
                //tn.y = ( tn.y + (int)(obs_map.y_size/2) ) % obs_map.y_size; // best if obs_map.y_size is even.
                tn.pole = -1;
                s->push_back(tn); c->push_back(1); // 1e-10
                //printf(" In south pole:");
                //tn.print();
                continue;
            }
            else PolesRemoved */ if (tn.y<0 && tn.x>=0 && tn.x<obs_map.x_size) {
                tn.y = obs_map.y_size - 1;
            }
            else if (tn.y>=obs_map.y_size && tn.x>=0 && tn.x<obs_map.x_size) {
                tn.y = 0;
            }
            else if ( tn.x>=0  &&  tn.x<obs_map.x_size  &&  tn.y>=0  &&  tn.y<obs_map.y_size ) {
                // do nothing
            }
            else
                continue;
            
		    s->push_back(tn);
		    
		    if(explicitMetricMat) {
		        csum = 0.0;
		        Dx[0] = (double)a; Dx[1] = (double)b;
		        for (int p=0; p<2; p++)
                    for (int q=0; q<2; q++) 
                        csum += (( metric_mat_map[p][q][n.x][n.y] + metric_mat_map[p][q][tn.x][tn.y] ) / 2.0) * Dx[p] * Dx[q];
                c->push_back(sqrt(csum));
		    }
		    else 
			    c->push_back( DistMap3x3[a+1][b+1] * sqrt( ( metric_map[n.x][n.y] + metric_map[tn.x][tn.y] ) / 2.0 ) );
			
			//tn.print();
		}
}

// ===========================================================
// Plotting functions
IplImage* ipl_image_p;
IplImage* ipl_image_p_processed;
CvSize map_size, opengl_size;
CvVideoWriter* writer;
char pressedKey;
// ------------

bool simPaused;
int userRobTargetX, userRobTargetY, userRobEvent;
void my_mouse_callback( int event, int x, int y, int flags, void* param ) {
    userRobTargetX = x;
    userRobTargetY = obs_map.y_size - y;
    if (event==CV_EVENT_RBUTTONDOWN || event==CV_EVENT_LBUTTONDOWN)
        userRobEvent = event; // left-click to engage, right-click to release
    int myuserrobot = power_voronoi_map[x][y];
    
    if (simPaused) {
        printf("$$$$ Mouse callback when paused: (%d, %d); tessellation belongs to rob %d.\n", x, y, myuserrobot);
        if (event==CV_EVENT_RBUTTONDOWN) {
            myNode n;
            n.x = x; n.y = obs_map.y_size - y; n.pole = 0;
            A_star_planner<myNode,double>::GraphNode_p thisGraphNode = planner->hash->getNodeInHash(n);
            while (thisGraphNode) {
                printf("(%d,%d)-", thisGraphNode->n.x, thisGraphNode->n.y);
                cvSetPixelColor(ipl_image_p, thisGraphNode->n.x, thisGraphNode->n.y, 255, 255, 255);
                thisGraphNode = thisGraphNode->came_from;
            }
            printf("\n\tplotted path!\n");
            cvShowImage("visualization", ipl_image_p);
        }
    }
}

// -----

void mycvPlotInit(void) {
    map_size.width = obs_map.x_size;
	map_size.height = obs_map.y_size;
    cvNamedWindow("visualization", CV_WINDOW_AUTOSIZE);
	ipl_image_p = cvCreateImage(map_size, IPL_DEPTH_8U, 3);
	ipl_image_p->origin = 1;
	ipl_image_p->dataOrder = IPL_DATA_ORDER_PIXEL;
	ipl_image_p->widthStep = ipl_image_p->width * 3;
	ipl_image_p_processed = cvCreateImage(map_size, IPL_DEPTH_8U, 3);
	ipl_image_p_processed->origin = 1;
	ipl_image_p_processed->dataOrder = IPL_DATA_ORDER_PIXEL;
	ipl_image_p_processed->widthStep = ipl_image_p->width * 3;
	
	if (allowusercontrol)
	    cvSetMouseCallback("visualization", my_mouse_callback, NULL);
	simPaused = false;
}

char envFname[1024];
bool takeScreenshots;

void mycvPlotFrame(int iter) {
    A_star_variables<double> thsnodeinfo;
    double weightIntensity, metricIntensity, testIntensity;
    myNode tempNode;
    bool isTesselationBoundary;
    int r, g, b;
    
    for (int a=0; a<obs_map.x_size; a++)
        for (int b=0; b<obs_map.y_size; b++) {
            if (obs_map[a][b])    cvSetPixelColor(ipl_image_p, a, b, 0, 0, 0);
            else if (a<POLE_CUTOFF || a>obs_map.x_size-POLE_CUTOFF)      cvSetPixelColor(ipl_image_p, a, b, 125, 125, 125);
            else {
                // Check if this is boundary of tessellation
                isTesselationBoundary = false;
                /*for (int j=a-1; j<=a+1; j++) {
                    for (int k=b-1; k<=b+1; k++) {
                        tempNode.x = j; tempNode.y = k; tempNode.pole = 0;
                        if (isAccessible(tempNode)  &&  power_voronoi_map[a][b]!=power_voronoi_map[j][k]) {
                            isTesselationBoundary = true;
                            break;
                        }
                    }
                    if (isTesselationBoundary) break;
                } */
                if (isTesselationBoundary)
                    cvSetPixelColor(ipl_image_p, a, b, 0, 0, 255);
                else {
                    if (explore)
                        weightIntensity = 1.0 + 1.55*weight_map[a][b];
                    else if (maxWeight-minWeight > 0.0)
                        weightIntensity = 1.0 + 1.55*(weight_map[a][b] - weightMinLim) / weightLimRange;
                    else
                        weightIntensity = 0.0;
                    //printf("%f:", weightIntensity);
                    
                    if (explore && altermetric)
                        metricIntensity = 1.0 + 1.55*metric_map[a][b];
                    else if (maxMetric-minMetric > 0.0) {
                        if (explicitMetricMat) 
                            metricIntensity = 1.0 + 1.55 * (metric_map[a][b] - minMetric) / (maxMetric - minMetric);
                        else 
                            metricIntensity = 1.0 + 1.55*(metric_map[a][b] - metricMinLim) / metricLimRange;
                        }
                    else
                        metricIntensity = 1.0;
                        
                    testIntensity = 0.0;
                    if (explore && smooth_entropy_map[a][b]>thresholdentropy) {
                        //printf("%d,%d,%d;", a, b, explore_contract_map[a][b]);
                        testIntensity = 2.5;
                    }
                        
                    // tessellation coloring
                    int rr = 50 + 60 * (((power_voronoi_map[a][b]+5)*7 % 3) + 1);
                    int gg = 50 + 25 * (((power_voronoi_map[a][b]+17)*5 % 7) + 1);
                    int bb = 50 + 40 * (((power_voronoi_map[a][b]+11)*2 % 5) + 1);
                    cvSetPixelColor(ipl_image_p, a, b, rr, gg, bb);
                }
            }
        }
    
    for (int a=0; a<hx.size(); a++) {
        cvGetPixelColor(ipl_image_p, hx[a], hy[a], &r, &g, &b);
        if (! (r==0 && g==0 && b==255) )
            cvSetPixelColor(ipl_image_p, hx[a], hy[a], 150, 150, 255);
    }
        
    for (int a=0; a<px.size(); a++) {
        cvSetPixelColor(ipl_image_p, px[a], py[a], 200, 100, 255, 2);
        if (userrobot==a)
            cvCircle( ipl_image_p, cvPoint(px[a], py[a]), 5*(int)sqrt(rr2[a]), CV_RGB(255,100,100), 1.5 );
        else
            cvCircle( ipl_image_p, cvPoint(px[a], py[a]), 5*(int)sqrt(rr2[a]), CV_RGB(255,255,255), 1.5 );
        char robnum[1024]; sprintf(robnum, "%d", a);
        CvFont font; cvInitFont(&font, CV_FONT_HERSHEY_PLAIN, 1.0, 1.0, 0, 1);
        //cvPutText( ipl_image_p, robnum, cvPoint(px[a]+2, py[a]+2), &font, cvScalar(200,200,255) );
        hx.push_back(px[a]); hy.push_back(py[a]);
    }
	cvShowImage("visualization", ipl_image_p);
	//cvConvertImage(ipl_image_p, ipl_image_p_processed, CV_CVTIMG_FLIP);
	//cvWriteFrame(writer, ipl_image_p_processed);
	
	takeScreenshots = false;
		
	for (int a=0; a<imgFrames.size(); a++) 
	    if (iter==imgFrames[a]) {
	        /*char imgFname[1024];
	        sprintf(imgFname, "env_out_files/frame_%d.png", iter);
	        cvSaveImage(imgFname, ipl_image_p); */
	        char imgFname[1024];
	        time_t ltime = time(NULL);
	        struct tm* Tm = localtime (&ltime);
	        //sprintf(timestamp, "_%d-%02d-%02d_%02d-%02d-%02d_.avi", Tm->tm_year+1900, Tm->tm_mon+1, Tm->tm_mday, Tm->tm_hour, Tm->tm_min, Tm->tm_sec);
            //sprintf(movFname, "env_out_files/%s%s", envFname, timestamp);
            sprintf(imgFname, "env_out_files/%s_%d-%02d-%02d_%02d-%02d-%02d_it-%d.png", envFname, Tm->tm_year+1900, Tm->tm_mon+1, Tm->tm_mday, Tm->tm_hour, Tm->tm_min, Tm->tm_sec, iter);
            cvSaveImage(imgFname, ipl_image_p);
            takeScreenshots = true;
	        break;
	    }
	
	pressedKey = cvWaitKey(10); // 50
	
	if (pressedKey=='p') {
	    simPaused = true;
	    //pthread_mutex_unlock( &mutexGlobal );
	    cvWaitKey();
	    //pthread_mutex_lock( &mutexGlobal );
	    pressedKey = -1;
	    simPaused = false;
	}
	else if (pressedKey=='c') {
	    char imgFname[1024];
	    time_t ltime = time(NULL);
	    struct tm* Tm = localtime (&ltime);
	    //sprintf(timestamp, "_%d-%02d-%02d_%02d-%02d-%02d_.avi", Tm->tm_year+1900, Tm->tm_mon+1, Tm->tm_mday, Tm->tm_hour, Tm->tm_min, Tm->tm_sec);
        //sprintf(movFname, "env_out_files/%s%s", envFname, timestamp);
        sprintf(imgFname, "env_out_files/%s_%d-%02d-%02d_%02d-%02d-%02d_it-%d.png", envFname, Tm->tm_year+1900, Tm->tm_mon+1, Tm->tm_mday, Tm->tm_hour, Tm->tm_min, Tm->tm_sec, iter);
        cvSaveImage(imgFname, ipl_image_p);
        pressedKey = -1;
        takeScreenshots = true;
	}
}

// ===========================================================

double PowerDist(double gVal, int RobIndex) {
    return (gVal*gVal - rr2[RobIndex]);
}

double heapFun(myNode& n, double g, double h, int s)
{
    return ( PowerDist(g, s) );
}

// ===========================================================
// OpenGL functions

/* void* StartGLdisplay(void*) {
    glutMainLoop();
}*/

void MainOpenGLIteration(void);

// -------------------------

static GLfloat light_ambient[]  = { 0.0, 0.0, 0.0, 1.0 };
static GLfloat light_diffuse[]  = { 0.5, 0.5, 0.5, 1.0 };
static GLfloat light_specular[] = { 0.2, 0.2, 0.2, 1.0 };
static GLfloat light_position[] = { 0.0, 0.0, -10.0, 0.0 };

static GLfloat mat_ambient[]    = { 0.7, 0.7, 0.7, 1.0 };
static GLfloat mat_diffuse[]    = { 0.8, 0.8, 0.8, 1.0 };
static GLfloat mat_specular[]   = { 1.0, 1.0, 1.0, 1.0 };
static GLfloat high_shininess[] = { 100.0 };

void myglPlotInit(int argc, char *argv[]) {
    glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE); //glutInitDisplayMode(GLUT_DEPTH | GLUT_DOUBLE | GLUT_RGBA);
	glutInitWindowSize(OPENGL_WIN_SIZE, OPENGL_WIN_SIZE);
	int win = glutCreateWindow("Sphere");

	glShadeModel(GL_SMOOTH);
	glClearColor(1.0f,1.0f,1.0f,0.0f);
	
	//glViewport(-2.0, -2.0, 4.0, 4.0);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	//glRotated(90.0, 1.0, 0.0, 0.0);
	gluPerspective(30.0, 1.0, 0.1, 100.0);
	//glScalef(1.0f, -1.0f, 1.0f);
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glScalef(-1.0f, 1.0f, 1.0f);
	gluLookAt (	0.0, 0.0, -5.0, // camera location
				0.0, 0.0, 0.0, // look at point
				0.0, 1.0, 0.0 ); // camera "up" direction

	glutDisplayFunc(MainOpenGLIteration);
	glutIdleFunc(MainOpenGLIteration);
	//glutKeyboardUpFunc(KeyDownHandlerOpenGL);
	
	zprReferencePoint[0] = 0.0; zprReferencePoint[1] = 0.0; zprReferencePoint[2] = 0.0; 
	zprInit();
	//zprSelectionFunc(drawAxes);     /* Selection mode draw function */
	//zprPickFunc(pick);              /* Pick event client callback   */
	
	// Setting up lights

	light_position[0] = 0.0; light_position[1] = 0.0; light_position[2] = -5.0; 
	glLightfv(GL_LIGHT0, GL_AMBIENT, light_ambient);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diffuse);
	glLightfv(GL_LIGHT0, GL_SPECULAR, light_specular);
	glLightfv(GL_LIGHT0, GL_POSITION, light_position);

	glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient);
	glMaterialfv(GL_FRONT, GL_DIFFUSE, mat_diffuse);
	glMaterialfv(GL_FRONT, GL_SPECULAR, mat_specular);
	glMaterialfv(GL_FRONT, GL_SHININESS, high_shininess);

	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glDepthFunc(GL_LESS);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_NORMALIZE);
	glEnable(GL_COLOR_MATERIAL);
}

// ===========================================================

void VideoCaptureFrame(void)
{
	//std::string movFname = CreateTimeStampedDerivedFilename(envFname, "avi");
	//double framerate = 20.0;
	
	//while (GLiterCount < 10) { printf(""); }
	
	//printf("Movie recording started: %s\n", movFname.c_str());
	//CvSize size;
	IplImage* glipl_image_p;
	IplImage* glipl_image_p_processed;
	
	//size.width = OPENGL_WIN_SIZE;
	//size.height = OPENGL_WIN_SIZE;
	//CvVideoWriter* writer;
	//writer = cvCreateVideoWriter ( movFname.c_str(), CV_FOURCC('D','I','V','X'), framerate, size);

	glipl_image_p_processed = cvCreateImage(opengl_size, IPL_DEPTH_8U, 3);
	glipl_image_p_processed->origin = 1;
	glipl_image_p_processed->widthStep = glipl_image_p_processed->width * 3;

	/*int framecount = 0;
	clock_t frameTime = clock();
	while (savemov)
	{
		size.width = WIDTH;
		size.height = HEIGHT; */
		glipl_image_p = cvCreateImageHeader(opengl_size, IPL_DEPTH_8U, 3);
		glipl_image_p->imageData = (char*)imageData;
		glipl_image_p->imageDataOrigin = glipl_image_p->imageData;
		glipl_image_p->widthStep = glipl_image_p->width * 3;
		glipl_image_p->origin = 1;
	
		cvConvertImage(glipl_image_p, glipl_image_p_processed, CV_CVTIMG_FLIP);
		cvWriteFrame(writer, glipl_image_p_processed);
		/* framecount++;
		
		while ( clock()-frameTime < CLOCKS_PER_SEC / framerate ) { printf(""); }
		frameTime = clock();
	} */
	
	//cvReleaseVideoWriter(&writer);
	//free(imageData);
	//printf("Movie (%d frames) saved in: %s\n", framecount, movFname.c_str());
	
	if (takeScreenshots) {
	    char imgFname[1024];
	    time_t ltime = time(NULL);
	    struct tm* Tm = localtime (&ltime);
	    //sprintf(timestamp, "_%d-%02d-%02d_%02d-%02d-%02d_.avi", Tm->tm_year+1900, Tm->tm_mon+1, Tm->tm_mday, Tm->tm_hour, Tm->tm_min, Tm->tm_sec);
        //sprintf(movFname, "env_out_files/%s%s", envFname, timestamp);
        sprintf(imgFname, "env_out_files/%s_3D_%d-%02d-%02d_%02d-%02d-%02d_it-%d.png", envFname, Tm->tm_year+1900, Tm->tm_mon+1, Tm->tm_mday, Tm->tm_hour, Tm->tm_min, Tm->tm_sec, iter);
        cvSaveImage(imgFname, glipl_image_p);
	}
}

// --------------------------------------------

    char movFname[1024], timestamp[1024];
    myNode tempNode;
    double thisDist, weightedIntegral, minIntegral;
    int tpx, tpy, robindex, ct, xx, yy;
    double ctrlmag, ctrlx, ctrly;
    double cxsum, cysum;
    double thiscellprob, thiscellentropy, thisSpeed;
    double delx, dely, r, s, z;
    bool robotincell, cellinsiderobotradius, abovethreshclosest;
    int tx, ty;
    // Graph
    GenericSearchGraphDescriptor<myNode,double> myGraph;
    

int main(int argc, char *argv[])
{
    strcpy(envFname, argv[1]);
    parseEnvFile(envFname);
    
    mycvPlotInit();
    // Movie setup
    //strcpy(movFname, envFname);
	time_t ltime = time(NULL);
	struct tm* Tm = localtime (&ltime);
	sprintf(timestamp, "_%d-%02d-%02d_%02d-%02d-%02d_.avi", Tm->tm_year+1900, Tm->tm_mon+1, Tm->tm_mday, Tm->tm_hour, Tm->tm_min, Tm->tm_sec);
	//strcat(movFname, timestamp);
    sprintf(movFname, "env_out_files/%s%s", envFname, timestamp);
    
    
    // Defining the unweighted transition costs
    DistMap3x3.init(3, 3, 0.0);
    double SQRT2 = sqrt(2.0);
    for (int a=-1; a<=1; a++)
		for (int b=-1; b<=1; b++) 
		    DistMap3x3[a+1][b+1] = sqrt((double)(a*a+b*b));
    
	// We describe the graph, cost function, heuristics, and the start & goal in this block
	myGraph.getHashBin_fp = &getHashBin;
	myGraph.getSuccessors_fp = &getSuccessors;
	myGraph.isAccessible_fp = &isAccessible;
	myGraph.TargetNode.x = -10; myGraph.TargetNode.y = -10; myGraph.TargetNode.pole = 0; // Inaccessible target
	// Set other variables
	myGraph.hashTableSize = obs_map.x_size + 11; // Since in this problem, "getHashBin" can return a max of value 201.
	myGraph.hashBinSizeIncreaseStep = obs_map.y_size; // By default it's 128. For this problem, we choose a higher value.
	
	power_voronoi_map.init(obs_map.x_size, obs_map.y_size, -1);
	controlIntegral.resize(px.size());
	controlAbsIntegral.resize(px.size());
	aboveThreshPtX.resize(px.size()); aboveThreshPtY.resize(px.size());
	geodesic_cotangents[0].resize(px.size()); geodesic_cotangents[1].resize(px.size());
	
	// OpenGL initialization and thread
	myglPlotInit(argc, argv);
	/* pthread_t glThread;
	pthread_attr_t glAttr;
	int glRC, glRet;
	//glRC = pthread_attr_init(&glAttr);
	//glRC = pthread_attr_setschedpolicy(&glAttr, SCHED_FIFO);
	//glRet = pthread_create( &glThread, NULL, StartGLdisplay, NULL );
	//
	*/
	
	WIDTH = glutGet(GLUT_WINDOW_WIDTH);
	HEIGHT = glutGet(GLUT_WINDOW_HEIGHT);
    opengl_size.width = WIDTH; opengl_size.height = HEIGHT;
	writer = cvCreateVideoWriter ( movFname, CV_FOURCC('D','I','V','X'), 10, opengl_size);
	
	
	//pthread_t vThread;
	//int vRet = pthread_create( &vThread, NULL, StartVideoCapturing, NULL );
	
	
	planner = new A_star_planner<myNode,double>;
	planner->setParams(1.0, 50, 0);
	userRobTargetX = -1; userRobTargetY = -1; userRobEvent = CV_EVENT_RBUTTONDOWN;
	pressedKey = -1;
	iter = 0;
	
	glutMainLoop();
    
    printf("Output written to %s.\n\n", movFname);
    cvReleaseVideoWriter(&writer);
    cvWaitKey();
}

// =======================

void MainOpenGLdraw(void) {
    // All drawings take place here
    //printf("inside draw\n");
    
    glClearDepth(1.0f);
	glClearColor(1.0f,1.0f,1.0f,0.0f);
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
	//printf("cleared image\n");
	
	// Draw axes
	glLineWidth(1.0);
	//printf("A\n");
	glDisable(GL_LIGHTING);
	//printf("B\n");
	for (int a=-1; a<=1; a++) {
	    //printf("C:%d\n", a);
		for (int b=0; b<3; b++)
		{
			std::vector<int> axisnum(3, 0);
			axisnum[b] = 1;
			glBegin(GL_LINES);
			glColor3d( 0.0, 0.0, 200.0 );
			glVertex3d ( 0.0, 0.0, 0.0 );
			glColor3d( 200.0, 200.0, 200.0 );
			glVertex3d ( a*4.0*axisnum[0], a*4.0*axisnum[1], a*4.0*axisnum[2] );
			glEnd();
		}
		//printf("D:%d\n", a);
    }
    //printf("E\n");
	glEnable(GL_LIGHTING);
	//printf("axes drawn... ");
	
	// draw the sphere
    double lati, longi, longip;
    int rr, gg, bb;
    for (int a=0; a<obs_map.x_size; a++) {
        //printf("\n%d (%d,%d): ", a, obs_map.x_size, obs_map.y_size); fflush(stdout);
        for (int b=0; b<obs_map.y_size; b++) {
            //printf("%d",b); fflush(stdout);
            lati = ((double)a)*dx - ox;
            longi = ((double)b)*dy - oy;
            if (b==obs_map.y_size)
                longip = 0.0;
            else
                longip = longi+dy;
            glBegin(GL_TRIANGLE_STRIP);
                cvGetPixelColor(ipl_image_p, a, b, &rr, &gg, &bb);
                if (a<POLE_CUTOFF || a>obs_map.x_size-POLE_CUTOFF)
                    glColor3d( 0.5, 0.5, 0.5 );
                else
                    glColor3d( ((double)rr)/255.0, ((double)gg)/255.0, ((double)bb)/255.0 );
                //glColor3d(0.9, 0.5, 0.5);
		        glNormal3d(sin(lati)*cos(longi), sin(lati)*sin(longi), cos(lati) );
		        glVertex3d(sin(lati)*cos(longi), sin(lati)*sin(longi), cos(lati) );
		        glVertex3d(sin(lati+dx)*cos(longi), sin(lati+dx)*sin(longi), cos(lati+dx) );
		        glVertex3d(sin(lati)*cos(longip), sin(lati)*sin(longip), cos(lati) );
		        glVertex3d(sin(lati+dx)*cos(longip), sin(lati+dx)*sin(longip), cos(lati+dx) );
	        glEnd();
        } 
    }
	
	
	// Done!
	//glutSwapBuffers();
}

// -----

void MainOpenGLIteration(void) {

    printf("\niter = %d\n", iter);
	    
	    /*if (explore) {
	        delete planner;
	        planner = new A_star_planner<myNode,double>;
	        planner->setParams(1.0, 50, 0);
	    }*/
	    
        if (explore) {
            // Update weight and metric maps based on sensor models - this discards information about weight, if provided, 
            // and metric if 'altermetric' is true
            //
            // Update probsum_map and probcount_map
            for (int x=0; x<unknownobs_map.x_size; x++)
                for (int y=0; y<unknownobs_map.y_size; y++) {
                    robotincell = false;
                    cellinsiderobotradius = false;
                    abovethreshclosest = false;
                    thiscellprob = probsum_map[x][y] / probcount_map[x][y];
                    thiscellentropy = -( thiscellprob*log(thiscellprob) + (1.0-thiscellprob)*log(1.0-thiscellprob) );
                    
                    for (int a=0; a<px.size(); a++) {
                        if (x==px[a] && y==py[a]) {
                            robotincell = true;
                            //robindex = a;
                        }
                        //if (x==aboveThreshPtX[a] && y==aboveThreshPtY[a])
                        //    abovethreshclosest = true;
                            
                        delx=x-px[a]; dely=y-py[a]; r=delx*delx+dely*dely;
                        if (r<=rr2[a]) {
                            cellinsiderobotradius = true;
                            if (thiscellentropy > freezeentropy) { // cell can be updated
                                // Generate randomized observation
                                s = 0.8 + 0.2 / exp(pow((double)r/(double)rr2[a], 10.0)); //1.0 - pow((double)r/(double)rr2[a], 10.0); // sensor model
                                z = (rand() < (int)((double)RAND_MAX*s) )? (double)unknownobs_map[x][y] : (double)(1-unknownobs_map[x][y]);
                                probsum_map[x][y] += z*s + thiscellprob*(1.0-s);
                                probcount_map[x][y]++;
                            }
                        }
                    }
                    
                    robindex = power_voronoi_map[x][y];
                    
                    //------------------------------------------------------
                    
                    entropy_map[x][y] = thiscellentropy;
                    // compute smoothened entropy
                    ct = 0;
                    smooth_entropy_map[x][y] = 0.0;
                    for (int a=-2; a<=2; a++)
                        for (int b=-entropysmoothenradius; b<=entropysmoothenradius; b++) {
                            xx = x + a; yy = y + b;
                            if (xx>=0  &&  xx<obs_map.x_size  &&  yy>=0  &&  yy<obs_map.y_size  &&  !obs_map[xx][yy]) {
                                smooth_entropy_map[x][y] += entropy_map[xx][yy];
                                ct++;
                            }
                        }
                    smooth_entropy_map[x][y] = smooth_entropy_map[x][y] / ct;
                    
                    //------------------------------------------------------
                    
                    // obstacle map
                    if (thiscellprob>obstacleprobthresh && !robotincell)
                        obs_map[x][y] = 1;
                    else
                        obs_map[x][y] = 0;
                        
                    // Weight map
                    if (thiscellentropy < thresholdentropy)
                        weight_map[x][y] = tinyweight; 
                    else
                        weight_map[x][y] = thiscellentropy/0.69315;
                    
                    if( x==aboveThreshPtX[robindex] && y==aboveThreshPtY[robindex] ) {
                        thisSpeed = sqrt(controlIntegral[robindex].x*controlIntegral[robindex].x+controlIntegral[robindex].y*controlIntegral[robindex].y); //controlAbsIntegral[robindex] / sqrt(controlIntegral[robindex].x*controlIntegral[robindex].x+controlIntegral[robindex].y*controlIntegral[robindex].y);
                        if (thisSpeed < controlmagzero ) { // &&  lowSpeedHistoryCount[robindex] > 10) { // TODO: change this absolute value
                            weight_map[x][y] = 100*controlAbsIntegral[robindex]; // TODO: change this value
                            printf("---rob: %d, weight at (%d,%d) set to %f\n", robindex, x, y, weight_map[x][y]);
                        }
                        /* else if (thisSpeed < 1e5)
                            lowSpeedHistoryCount[robindex]++;
                        else
                            lowSpeedHistoryCount[robindex] = __max(0, lowSpeedHistoryCount[robindex]-1); */
                    }
                        
                    // Metric map (assumes not a matrix-based metric map)
                    if (altermetric) {
                        if (thiscellentropy < thresholdentropy) {
                            /*if (cellinsiderobotradius) {
                                metric_map[x][y] = tinymetric;
                                
                            }*/
                            //if( explore_contract_map[x][y] )
                            //    metric_map[x][y] = tinymetric;
                            //else
                                metric_map[x][y] = tinymetric; //thresholdentropy; // tinymetric*100.0;
                        }
                        else
                            metric_map[x][y] = thiscellentropy/0.69315; //thiscellentropy;
                    }
                    //    metric_map[x][y] = __max(thiscellentropy, thresholdentropy);
            }
            
            // Smoothen entropy
            
            
        }
	    
	    if (explicitMetricMat) {
	        // Precomputations of geodesic_cotangents[i][n][e] - i=0,1 ; e=0..7; n=0...(N-1)
	        int ppx, ppy;
	        
	        for (int n=0; n<px.size(); n++) {
	            geodesic_cotangents[0][n].clear(); geodesic_cotangents[1][n].clear();
	            //printf("metricmat: [%f %f; %f %f]\n", metric_mat_map[0][0][px[n]][py[n]], metric_mat_map[0][1][px[n]][py[n]], metric_mat_map[1][0][px[n]][py[n]], metric_mat_map[1][1][px[n]][py[n]]);
	            
	            // iterate over the edges
	            for (int a=-1; a<=1; a++)
		            for (int b=-1; b<=1; b++) {
		                if (a==0 && b==0) continue;
		                
		                csum = 0.0; cxsum = 0.0; cysum = 0.0;
		                Dx[0] = (double)a; Dx[1] = (double)b;
		                for (int p=0; p<2; p++) {
		                    ppx = (px[n]+a<0)? px[n] : ((px[n]+a>=obs_map.x_size)? px[n] : px[n]+a);
		                    ppy = (py[n]+b<0)? py[n] : ((py[n]+b>=obs_map.y_size)? py[n] : py[n]+b);
	                        for (int q=0; q<2; q++) 
	                            csum += (( metric_mat_map[p][q][px[n]][py[n]] + metric_mat_map[p][q][ppx][ppy] ) / 2.0) * Dx[p] * Dx[q];
	                        cxsum += (( metric_mat_map[p][0][px[n]][py[n]] + metric_mat_map[p][0][ppx][ppy] ) / 2.0) * Dx[p];
	                        cysum += (( metric_mat_map[p][1][px[n]][py[n]] + metric_mat_map[p][1][ppx][ppy] ) / 2.0) * Dx[p];
	                    }
	                    
	                    //c->push_back(sqrt(csum));
	                    geodesic_cotangents[0][n].push_back(cxsum/sqrt(csum));
	                    geodesic_cotangents[1][n].push_back(cysum/sqrt(csum));
		            }
		        
		    }
	    }

        expandCount1 = 0; expandCountTot = 0;
        explore_contract_map.reset(0);
	    // Compute tesellations
	    power_voronoi_map.reset(-1);
	    //myGraph.isAccessible_fp = NULL; // This is taken care by 'getSuccessors'
	    myGraph.SeedNodes.clear();
	    for (int a=0; a<px.size(); a++) {
	        tempNode.x = px[a]; tempNode.y = py[a]; tempNode.pole = 0;
	        tempNode.SeedNeighbourX = 0; tempNode.SeedNeighbourY = 0; 
	        tempNode.SeedIndex = a;
	        myGraph.SeedNodes.push_back(tempNode);
	        controlIntegral[a] = createPtXY();
	        controlAbsIntegral[a] = 0.0;
	        if (explore) {
	            aboveThreshPtX[a] = -1; aboveThreshPtY[a] = -1;
	            explore_contract_map[px[a]][py[a]] = 1;
	        }
	        //printf("Robot %d: (%d,%d)-%d; ", a, px[a], py[a], isAccessible(myGraph.SeedNodes[a]));
	    }
	    planner->heapFun_fp = heapFun;
	    planner->event_NodeExpanded_g = event_NodeExpanded;
	    planner->event_SuccUpdated_g = event_SuccUpdated;
	    /*if (explore)
	        planner->init(&myGraph);
	    else */
    	    planner->clearLastPlanAndInit(&myGraph);
	    printf("Now computing tesellations...\n");
	    planner->plan();
	    for (int i=0; i<obs_map.x_size; i++)
            for (int j=0; j<obs_map.y_size; j++) {
                tempNode.x = i; tempNode.y = j; tempNode.pole = 0;
                if (isAccessible(tempNode)) 
                    power_voronoi_map[i][j] = planner->getNodeInfo(tempNode).seedLineage;
            }
	    
	    
	    // User control
	    if (allowusercontrol) {
            if (userRobEvent==CV_EVENT_LBUTTONDOWN) {
                userrobot = power_voronoi_map[userRobTargetX][userRobTargetY];
                printf("$$$$ Mouse location: (%d, %d); User controlling rob %d\n", userRobTargetX, userRobTargetY, userrobot);
            }
            else {
                userrobot = -1;
                printf("$$$$ Mouse location: (%d, %d); No robot engaged.\n", userRobTargetX, userRobTargetY);
            }
        }
	    
	    // Graphics
	    printf("Now plotting 2D image...\n");
	    mycvPlotFrame(iter);
	    printf("Now plotting 3D image...\n");
	    MainOpenGLdraw();
	    printf("3D image drawn!\n");
        
        // evaluate next position
        for (int a=0; a<px.size(); a++) {
            if (allowusercontrol && userrobot==a) {
                printf("+ robot %d> pos: (%d,%d); User-specified target: (%d, %d).\n", a, px[a], py[a], userRobTargetX, userRobTargetY);
                ctrlx = (double)(userRobTargetX-px[a]);
                ctrly = (double)(userRobTargetY-py[a]);
                ctrlmag = sqrt(ctrlx*ctrlx + ctrly*ctrly);
                ctrlx = ctrlx/ctrlmag; ctrly = ctrly/ctrlmag;
                //printf(" ctrl: (%f, %f) \n", ctrlx, ctrly);
                tx = px[a] + ( (ctrlx>CTRLLIM)? 1 : ( (ctrlx<-CTRLLIM)? -1 : 0 ) );
                ty = py[a] + ( (ctrly>CTRLLIM)? 1 : ( (ctrly<-CTRLLIM)? -1 : 0 ) );
                //printf("--- %d,%d ; %d,%d", tx, ty, px[a], py[a]);
                if ( ( tx>=0  &&  tx<obs_map.x_size  &&  ty>=0  &&  ty<obs_map.y_size   &&   !obs_map[tx][ty] ) ) {
                    px[a] = tx; py[a] = ty;
                }
                else if ( ( tx>=0  &&  tx<obs_map.x_size  &&  py[a]>=0  &&  py[a]<obs_map.y_size   &&   !obs_map[tx][py[a]] ) ) {
                    px[a] = tx;
                }
                else if ( ( px[a]>=0  &&  px[a]<obs_map.x_size  &&  ty>=0  &&  ty<obs_map.y_size   &&   !obs_map[px[a]][ty] ) ) {
                    py[a] = ty;
                }
            }
            else {
                ctrlmag = sqrt( pow(controlIntegral[a].y, 2.0) + pow(controlIntegral[a].x, 2.0) );
                
                printf("+ robot %d> 'expandCount1:%d,expandCountTot:%d' pos: (%d,%d); control: (%f, %f) %f/%f; Absoulte integral: %f; Above thresh pt: (%d,%d), ", a, expandCount1,expandCountTot, px[a], py[a], controlIntegral[a].x, controlIntegral[a].y, ctrlmag, controlmagzero, controlAbsIntegral[a], aboveThreshPtX[a], aboveThreshPtY[a] );
                if (aboveThreshPtX[a]>0)
                    printf("pt belongs to rob %d\n", power_voronoi_map[aboveThreshPtX[a]][aboveThreshPtY[a]] );
                else
                    printf("pt belongs to rob n/a\n");
                
                //ctrlmag = sqrt( pow(controlIntegral[a].y, 2.0) + pow(controlIntegral[a].x, 2.0) );
                if ( (explore && ( ctrlmag>=controlmagzero || (ctrlmag>=0.00001 && aboveThreshPtX[a]<0) ) ) 
                        || (!explore && ctrlmag>=0.00001) ) { // 0.00001. TODO: Change the abs value of 1e5
                    ctrlx = controlIntegral[a].x / ctrlmag;
                    ctrly = controlIntegral[a].y / ctrlmag;
                    //printf(" ctrl: (%f, %f) \n", ctrlx, ctrly);
                    delx = (ctrlx>CTRLLIM)? 1 : ( (ctrlx<-CTRLLIM)? -1 : 0 );
                    dely = (ctrly>CTRLLIM)? 1 : ( (ctrly<-CTRLLIM)? -1 : 0 );
                    
                    /* if (px[a]==0 && delx==-1) {
                        py[a] = ( py[a] + (int)(obs_map.y_size/2) ) % obs_map.y_size;
                    }
                    else if (px[a]==(obs_map.x_size-1) && delx==1) {
                        py[a] = ( py[a] + (int)(obs_map.y_size/2) ) % obs_map.y_size;
                    } */
                    if (px[a]==POLE_CUTOFF && delx==-1) {
                        py[a] = ( py[a] + (int)(obs_map.y_size/2) ) % obs_map.y_size;
                    }
                    else if (px[a]==(obs_map.x_size-1-POLE_CUTOFF) && delx==1) {
                        py[a] = ( py[a] + (int)(obs_map.y_size/2) ) % obs_map.y_size;
                    }
                    else if (py[a]==0 && dely==-1) {
                        px[a] += delx;
                        py[a] = obs_map.y_size - 1;
                    }
                    else if (py[a]==(obs_map.y_size-1) && dely==1) {
                        px[a] += delx;
                        py[a] = 0;
                    }
                    else {
                        px[a] += delx;
                        py[a] += dely;
                    }
                    
                }
            }
        }

    iter++;
    
    // save movie frame
	if (!imageData)
	{
		int BIT_R = sizeof(unsigned char);
		int BIT_G = sizeof(unsigned char);
		int BIT_B = sizeof(unsigned char);
		imageData = malloc(WIDTH*HEIGHT*(BIT_R+BIT_G+BIT_B)+1000);
	}
	glReadPixels(0, 0, WIDTH, HEIGHT, GL_BGR, GL_UNSIGNED_BYTE, imageData);
	VideoCaptureFrame();
	
    
    glutSwapBuffers();
}
